/*
    ext2online.c -- resizes a mounted ext2 fs
    Copyright (C) 1999 Andreas Dilger <adilger@enel.ucalgary.ca>
    Copyright (C) 1999 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This program is to be used in cooperation with Andreas Dilger's
  online resize patch for the linux kernel.
*/

static const char _ext2online_c[] = "$Id: ext2online.c,v 1.4 1999/08/22 20:35:08 buytenh Exp $";

#include "config.h"

#ifndef HAVE_LINUX_EXT2_FS_H
#error This program only works on linux
#endif

#ifndef MNTTYPE_EXT2
#define MNTTYPE_EXT2 "ext2"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <mntent.h>
#include <sys/mount.h>
#include "ext2.h"

char *prog;

void usage(FILE *outfile)
{
	fprintf(outfile, "ext2online v0.1 - 08/11/1999 for EXT2FS %s\n",
		EXT2FS_VERSION);
	fprintf(outfile, "usage: %s device [new_size[bkmgt]]\n", prog);
	fprintf(outfile,
"       where new_size is in ext2 filesystem blocks (1k, 2k, or 4k),\n"
"       disk Blocks (512), Kilo-, Mega-, Giga-, or Terabytes\n");
}

void ext2_parsedev(char *filename, char **dev, char **dir, char **type)
{
	FILE		*tabfile;
	struct mntent	*ent;

	tabfile = setmntent(filename, "r");

	while  ((ent = getmntent(tabfile)) != NULL)
	{
		if (strcmp(*dev, ent->mnt_fsname) == 0)
		{
			if (strcmp(ent->mnt_type, MNTTYPE_EXT2) != 0)
				break;
			*dir = (char *)malloc(strlen(ent->mnt_dir) + 1);
			if (*dir == NULL)
				*dir = ent->mnt_dir;
			else
				strcpy(*dir, ent->mnt_dir);
			break;
		}
		else if (strcmp(*dev, ent->mnt_dir) == 0)
		{
			if (strcmp(ent->mnt_type, MNTTYPE_EXT2) != 0)
				break;
			*dir = *dev;
			*dev = (char *)malloc(strlen(ent->mnt_fsname) + 1);
			if (*dev == NULL)
				*dev = ent->mnt_dir;
			else
				strcpy(*dev, ent->mnt_fsname);
			break;
		}
	}
	if (*dir != NULL && ent != NULL)
	{
		*type = (char *)malloc(strlen(ent->mnt_type) + 1);
		if (*type == NULL)
			*type = ent->mnt_type;
		else
			strcpy(*type, ent->mnt_type);
	}
	endmntent(tabfile);
}

int find_in_inode(struct ext2_fs *fs, struct ext2_inode *inode, blk_t block)
{
	if (inode->i_mode)
	{
		int			 i;
		int			 count = 0;
		struct ext2_buffer_head	*bh;
		__u32			*udata;
		int			 u32perblock = fs->blocksize >> 2;

		for (i = 0; i < EXT2_NDIR_BLOCKS; i++)
			if (inode->i_block[i] == block)
				return i + 1;
			else if (inode->i_block[i])
				count += 2;

		if (!inode->i_block[EXT2_IND_BLOCK] || count >= inode->i_blocks)
			return 0;

		bh = ext2_bread(fs, inode->i_block[EXT2_IND_BLOCK]);
		udata = (__u32 *)bh->data;
		for (i = 0; i < u32perblock && count < inode->i_blocks; i++)
			if (udata[i] == block)
			{
				ext2_brelse(bh, 0);
				return i + EXT2_NDIR_BLOCKS + 1;
			}
			else if (udata[i])
				count += 2;

		ext2_brelse(bh, 0);
		if (!inode->i_block[EXT2_DIND_BLOCK] ||
		    count >= inode->i_blocks)
			return 0;

		bh = ext2_bread(fs, inode->i_block[EXT2_DIND_BLOCK]);
		udata = (__u32 *)bh->data;
		for (i = 0; i < u32perblock && count < inode->i_blocks; i++)
		{
			struct ext2_buffer_head	*bh2;
			__u32			*udata2;
			int			 j;

			if (udata[i])
				bh2 = ext2_bread(fs, udata[i]);
			udata2 = (__u32 *)bh2->data;
			for (j = 0; j < u32perblock && count < inode->i_blocks;
			     j++)
			{
				if (udata2[j] == block)
				{
					ext2_brelse(bh, 0);
					ext2_brelse(bh2, 0);
					return (i + 1) * u32perblock + j +
					       EXT2_NDIR_BLOCKS + 1;
				}
				else if (udata2[j])
					count += 2;
			}
			ext2_brelse(bh2, 0);
		}
		ext2_brelse(bh, 0);
	}
	return 0;
}

int ext2_check_gdblocks(struct ext2_fs *fs, struct ext2_inode *inode,
			blk_t newgdblocks)
{
	int	group;
	blk_t	block;
	blk_t	gdb;

	for (group = 0, gdb = fs->sb.s_first_data_block + 1;
	     group < fs->numgroups; group++, gdb += fs->sb.s_blocks_per_group)
	for (block = fs->gdblocks; block < newgdblocks; block++)
	{
		if (fs->opt_debug)
			fprintf(stderr,
				"checking for group block %d in Bond\n", gdb);

		if (ext2_is_group_sparse(fs, group) &&
		    !find_in_inode(fs, inode, gdb + block))
		{
			fprintf(stderr, "%s: group %d, block %u not reserved\n",
				prog, group, gdb + block);
			return 0;
		}
	}
	return 1;
}

blk_t ext2_make_group(struct ext2_fs *fs, int group, blk_t blocks, blk_t skip)
{
	struct ext2_buffer_head	*bh;
	struct ext2_group_desc	*gdp;
	blk_t			 start;
	blk_t			 bboffset, iboffset, itoffset;
	int			 max_inodes;
	__u32			*udata;
	int			 i;

	if (fs->opt_verbose)
		fprintf(stderr,
			"%s: creating group %d with %u blocks\n",
			prog,
			group,
			blocks);

	if (blocks < fs->adminblocks + 50)
		return 0;

	bh = ext2_bread(fs, fs->sb.s_first_data_block +
			howmany(group * sizeof(struct ext2_group_desc),
				fs->blocksize));
	bh->dirty = 1;
	gdp = (struct ext2_group_desc *)bh->data +
	      (group % sizeof(struct ext2_group_desc));

	start = fs->sb.s_first_data_block + group * fs->sb.s_blocks_per_group;

#if 0
	/* FIXME: need to make sure that we don't overlap metadata */
	skip = skip * group % fs->sb.s_blocks_per_group;
#endif

	/* FIXME: need to handle skip at some point too. */
	itoffset = fs->itoffset + skip;
	if (itoffset + fs->inodeblocks > blocks)
	{
		fprintf(stderr, "%s: resetting metadata offset for group %d\n",
			prog, group);
		itoffset = 3 + ext2_is_group_sparse(fs, group)?fs->gdblocks : 0;
	}
	itoffset += start;

	/* Mark metadata blocks used in block bitmap */
	if (ext2_is_group_sparse(fs, group))
	{
		if (fs->opt_debug)
		{
			fprintf(stderr,
				"clear/mark superblock "
				"%u used\n",
				start);

			fprintf(stderr,
				"clear/mark group descriptors "
				"%u %u used\n",
				start, start + fs->gdblocks - 1);
		}

		bboffset = itoffset - 2;
		ext2_set_block_state(fs, start, 1, 0);
		for (i = 1; i <= fs->gdblocks; i++)
			ext2_set_block_state(fs, start + i, 1, 0);
		gdp->bg_free_blocks_count = 0;
	}
	else
	{
		bboffset = start;
		gdp->bg_free_blocks_count = 1;
	}
	iboffset = bboffset + 1;
	gdp->bg_block_bitmap = bboffset;
	gdp->bg_inode_bitmap = iboffset;
	gdp->bg_inode_table  = itoffset;
	gdp->bg_free_blocks_count += blocks - fs->adminblocks;
	gdp->bg_free_inodes_count = fs->sb.s_inodes_per_group;
	gdp->bg_used_dirs_count = 0;
	gdp->bg_pad = 0;
	gdp->bg_reserved[0] = gdp->bg_reserved[1] = gdp->bg_reserved[2] = 0;

	if (fs->opt_debug)
	{
		fprintf(stderr, "clear/mark block bitmap %u used\n", bboffset);
		fprintf(stderr, "clear/mark inode bitmap %u used\n", iboffset);
		fprintf(stderr, "clear/mark inode blocks %u through %u used\n",
			itoffset, itoffset + fs->inodeblocks - 1);
	}

	ext2_set_block_state(fs, bboffset, 1, 0);
	ext2_set_block_state(fs, bboffset, 1, 0);
	ext2_zero_blocks(fs, bboffset, 1);
	ext2_set_block_state(fs, iboffset, 1, 0);
	ext2_zero_blocks(fs, iboffset, 1);
	for (i = 0; i < fs->inodeblocks; i++)
		ext2_set_block_state(fs, itoffset + i, 1, 0);
	ext2_zero_blocks(fs, itoffset, fs->inodeblocks);
	ext2_brelse(bh, 0);

	bh = ext2_bread(fs, iboffset);
	bh->dirty = 1;
	udata = (__u32 *)bh->data;

	/* Mark unavailable blocks at end of group used */
	if (fs->opt_verbose && blocks != fs->sb.s_blocks_per_group)
		fprintf(stderr, "%s: mark end blocks %u through %u used\n",
			prog, start+blocks, start+fs->sb.s_blocks_per_group-1);

	for (i = blocks; i < fs->sb.s_blocks_per_group; i++)
		udata[i>>3] |= _bitmap[i&7];
	ext2_brelse(bh, 0);

	/* Mark unavailable inodes at end of table used */
	bh = ext2_bread(fs, iboffset);
	bh->dirty = 1;
	max_inodes = fs->blocksize / sizeof(struct ext2_inode) *
		     fs->inodeblocks;
	udata = (__u32 *)bh->data;

	if (fs->opt_verbose && max_inodes != fs->sb.s_inodes_per_group)
		fprintf(stderr, "%s: mark inodes %u through %u used\n",
			prog, fs->sb.s_inodes_per_group, max_inodes);

	for (i = fs->sb.s_inodes_per_group; i < max_inodes; i++)
		udata[i>>3] |= _bitmap[i&7];
	ext2_brelse(bh, 0);

	return blocks;
}

blk_t ext2_online_primary(struct ext2_fs *fs, blk_t newsize)
{
	struct ext2_inode	 inode;
	int			 group;
	int			 newgroups;
	int			 newgdblocks;
	blk_t			 skip = 0;
	blk_t			 realsize;

	if (fs->opt_verbose)
		fprintf(stderr, "%s: resize to %u blocks\n", prog, newsize);

	if (newsize <= fs->sb.s_blocks_count)
		return newsize;

	newgroups = howmany(newsize - fs->sb.s_first_data_block, fs->sb.s_blocks_per_group);
	newgdblocks = howmany(newgroups * sizeof(struct ext2_group_desc), fs->blocksize);

	if (fs->opt_debug)
	{
		fprintf(stderr,
			"%d old groups, %d blocks\n",
			fs->numgroups,
			fs->gdblocks);

		fprintf(stderr,
			"%d new groups, %d blocks\n",
			newgroups,
			newgdblocks);
	}

	/* We need to use more GDT blocks */
	if (newgdblocks > fs->gdblocks)
	{
		ext2_read_inode(fs, EXT2_RESIZE_INO, &inode);
		if (!ext2_check_gdblocks(fs, &inode, newgdblocks))
			return 0;
	}

#if 0
	if (fs->numgroups > 2)	/* Try to keep RAID spacings */
	{
		skip = (fs->gd[2].bg_block_bitmap - fs->gd[1].bg_block_bitmap) %
		       fs->sb.s_blocks_per_group;
		if (skip != (fs->gd[2].bg_inode_bitmap - fs->gd[1].bg_inode_bitmap) %
			    fs->sb.s_blocks_per_group)
			skip = 0;
	}
#endif

	/* Fill existing group */
	realsize = min(fs->numgroups * fs->sb.s_blocks_per_group +
			fs->sb.s_first_data_block, newsize);

	if (fs->opt_debug && realsize > fs->sb.s_blocks_count)
		fprintf(stderr, "filling last group to %u blocks\n", realsize);

	/* Add new groups */
	for (group = fs->numgroups; group < newgroups; group++)
	{
		blk_t blocks;

		blocks = min(fs->sb.s_blocks_per_group, newsize - realsize);
		realsize += ext2_make_group(fs, group, blocks, skip);
	}

	return realsize;
}

void ext2_online_backups(struct ext2_fs *fs, int groups)
{
}

int main(int argc, char *argv[])
{
	char			*dev, *type, *dir = NULL;
	struct ext2_dev_handle	*handle;
	struct ext2_fs		*fs;
	int		         newgroups;
	blk_t		         newsize, devsize, realsize;
	char			 mod = '\0';
	char			 resize[20];

	prog = argv[0];
	dev = argv[1];

	if (argc < 2 || argc > 3)
	{
		usage(stderr);
		return 1;
	}

	ext2_parsedev(MOUNTED, &dev, &dir, &type);
	if (dir == NULL)
	{
		/* Maybe we should look through MNTTAB as well? */
		fprintf(stderr, "%s: can't find %s, is it mounted?\n",
			prog, dev);
		usage(stderr);
		return 1;
	}

	handle = ext2_make_dev_handle_from_file(dev);
	if (handle == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, prog);
		return 1;
	}

	if (argc == 3)
	{
		sscanf(argv[2], "%i%c", &newsize, &mod);

		switch (mod)
		{
			case 't':
			case 'T': newsize *= 1024;
			case 'g':
			case 'G': newsize *= 1024;
			case 'm':
			case 'M': newsize *= 1024;
			case 'k':
			case 'K': /* Need to load fs to find block size */
			case  0 :
			case 'b':
			case 'B': /* do nothing */ break;
			default: usage(stderr);
		}

		newgroups = howmany(newsize, 1024); /* worst case 1k blocks */

		if (ext2_max_groups < newgroups)
			ext2_max_groups = newgroups;
	}

	if ((fs = ext2_open(handle, EXT2_VALID_FS)) == NULL)
	{
		fprintf(stderr, "%s: can't open %s\n", dev, prog);
		return 1;
	}

	if (mod != '\0' && mod != 'b' && mod != 'B')
	{
		newsize /= fs->blocksize / 1024;
	}

	devsize = fs->devhandle->ops->get_size(fs->devhandle->cookie);

	if (argc == 2)	/* Resize to the full size if none is given */
		newsize = devsize;
	else if (devsize < newsize)
	{
		fprintf(stderr, "%s: %s has %u blocks not %u, resize aborted\n",
			prog, dev, devsize, newsize);
		return 1;
	}

	if (newsize >= 1<<28 && fs->blocksize == 1024)
	{
		fprintf(stderr, "%s: can't resize 1k block filesystem over 256GB.\n", prog);
		fprintf(stderr, "%s: too much space would be wasted, sorry.\n", prog);
		return 1;
	}

	if ((realsize = ext2_online_primary(fs, newsize)) == 0)
	{
		fprintf(stderr, "%s: can't resize past %u blocks\n",
			prog, fs->gdblocks * fs->blocksize /
			sizeof(struct ext2_group_desc));
		return 1;
	}
	else if (realsize == devsize)
	{
		fprintf(stderr, "%s: unable to resize %s\n", prog, dev);
		return 1;
	}
	else if (realsize < newsize)
		fprintf(stderr, "%s: resizing to %u blocks\n", prog, realsize);

	sprintf(resize, "resize=%u", realsize);

	if (fs->opt_debug)
		fprintf(stderr,
			"Calling mount() for %s(%s) with %s\n",
			dev,
			dir,
			resize);

	sync(); sync();
	mount(dev, dir, type, MS_MGC_VAL | MS_REMOUNT, resize);

	ext2_online_backups(fs, newgroups);
	ext2_close(fs);

	return 0;
}

